#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from .lexer import Lexer


class GetoptLexer(Lexer):
    def __init__(self, lexer, known_commands, known_options=None, aliases=None):
        self._lexer = lexer
        self._current_token = 0
        self._known_commands = known_commands
        self._known_options = known_options or {}
        self._aliases = aliases or {}

    def input(self, text):
        self._lexer.input(text)
        self._current_token = 0

    def token(self):
        token = self._lexer.token()
        if token is not None:
            if self._current_token == 0:
                translated_token = self._translate_alias(token.value)
                if translated_token in self._known_commands:
                    token.type = 'COMMAND_{}'.format(translated_token.upper().replace('-', '_'))
                else:
                    token.type = 'COMMAND'
            else:
                if token.value in self._known_options:
                    token.type = "OPT{}".format(token.value.upper().replace('-', '_'))
            self._current_token += 1
        return token

    def get_position(self):
        return self._lexer.get_position()

    def _translate_alias(self, token):
        return self._aliases.get(token, token)



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from abc import abstractmethod, ABCMeta
from ply import yacc
from .tablexer import TabLexer


class CompleterLang():
    """Class encapsulating a language (or grammar) used by tab completion

    Derived classes should define their ply.yacc rules in their body, which is
    then translated at instantiation time.
    """
    __metaclass__ = ABCMeta

    def __init__(self):
        self._initialize_rules()
        self._parser = yacc.yacc(module=self, write_tables=False, debug=False, errorlog=yacc.NullLogger())
        self._lexer = TabLexer(self._construct_lexer())
        self._expected_tokens = []
        self._token_position = -1

    @abstractmethod
    def _construct_lexer(self):
        raise NotImplementedError

    def _initialize_rules(self):
        pass

    def get_expected_tokens(self, text, drop_last_token):
        self._lexer.set_drop_last_token(drop_last_token)
        self._parser.parse(text, lexer=self._lexer)
        return self._expected_tokens, self._lexer.get_replaced_token(), self._token_position

    def p_error(self, p):
        if p is None:
            # EOF
            return
        elif p.type == 'TAB':
            parser_state = self._parser.state

            # now handle the error that the TAB token caused
            self._token_position = p.lexpos
            self._collect_expected_tokens_and_productions(parser_state)
            self._parser.errok()

    def _collect_expected_tokens_and_productions(self, parser_state):
        self._expected_tokens = []
        self._collect_expected_tokens_for_a_given_state(parser_state)

    def _collect_expected_tokens_for_a_given_state(self, state):
        for token, next_state in self._iter_parser_actions(state):
            if token != '$end':
                self._expected_tokens.append(token)

            next_state = self._shift_production_if_needed(next_state, token)

            # negative next_state value means one of two things:
            #
            #    1) there's a syntax error and the current token does not match
            #       the rule we are evaluating right now
            #
            #    2) this token would trigger another rule shift, which we don't
            #       support right now.
            #

            if next_state >= 0:
                self._collect_expected_productions(next_state)

    def _collect_expected_productions(self, state):
        for _, next_state in self._iter_parser_actions(state):
            if next_state < 0:
                # production shift, we care about production shifts which would translate the
                # next_state token
                production = self._lookup_production(next_state)
                if production.len and production.name not in self._expected_tokens:
                    self._expected_tokens.append(production.name)

    def _iter_parser_actions(self, parser_state):
        return self._parser.action[parser_state].items()

    @staticmethod
    def _are_we_shifting_a_production(state):
        return state < 0

    def _shift_production(self, production, token):
        translated_state = self._get_target_state_after_shifting_production(production)
        try:
            # this might be negative (e.g. indicate a production shift).
            return self._parser.action[self._parser.goto[translated_state][production.name]][token]
        except KeyError:
            # this indicates a syntax error, the current token
            # does not match any rules where this production
            # is referenced from
            return -1

    def _shift_production_if_needed(self, state, token):
        if self._are_we_shifting_a_production(state):
            production = self._lookup_production(state)
            return self._shift_production(production, token)
        return state

    def _lookup_production(self, state):
        return self._parser.productions[-state]

    def _get_target_state_after_shifting_production(self, production):
        return self._parser.statestack[-production.len - 1]



#############################################################################
# Copyright (c) 2023 One Identity LLC.
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=unused-import,function-redefined,useless-parent-delegation

try:
    from _syslogng import LogSource, LogFetcher, LogFetcherResult
    from _syslogng import InstantAckTracker, ConsecutiveAckTracker, BatchedAckTracker

except ImportError:
    import warnings
    from enum import Enum, auto

    warnings.warn("You have imported the syslogng package outside of syslog-ng, "
                  "thus some of the functionality is not available. "
                  "Defining fake classes for those exported by the underlying syslog-ng code")

    # Fake LogSource
    class LogSource(dict):
        def __init__(self):
            pass

    # Fake LogFetcher
    class LogFetcher(dict):
        def __init__(self):
            pass

    class LogFetcherResult(Enum):
        ERROR = auto()
        NOT_CONNECTED = auto()
        SUCCESS = auto()
        TRY_AGAIN = auto()
        NO_DATA = auto()

    # Fake InstantAckTracker
    class InstantAckTracker(dict):
        def __init__(self, ack_callback):
            self.ack_callback = ack_callback

    # Fake ConsecutiveAckTracker
    class ConsecutiveAckTracker(dict):
        def __init__(self, ack_callback):
            self.ack_callback = ack_callback

    # Fake BatchedAckTracker
    class BatchedAckTracker(dict):
        def __init__(self, timeout, batch_size, batched_ack_callback):
            self.timeout = timeout
            self.batch_size = batch_size
            self.batched_ack_callback = batched_ack_callback


class LogSource(LogSource):
    """Base class for building an asynchronous source driver

    A LogSource is a base class that allows the implementation of a
    syslog-ng source driver.  Derive your source implementation from this
    class and override the run() method to add your specific behaviour.

    Internally, this class starts a dedicated thread and allows you to
    use an asynchronous framework such as asyncio to produce messages.

    It easily maps to any source mechanism that pushes messages to syslog-ng
    (e.g.  queueing protocols such as mqtt or HTTP server implementations)
    """
    flags = {}
    parse_options = None

    def __init__(self):
        """Constructs this LogSource instance
        """
        super().__init__()

    def init(self, options):
        """Initialize this LogSource instance

        Returns:
            bool: True to indicate success
        """
        return True

    def deinit(self):
        """Deinitialize this LogSource instance
        """

    def run(self):
        """Run the main loop for the source

        This method is invoked by syslog-ng to start producing messages in
        an asynchronous manner.  This function is executed in a dedicated
        thread and should run until request_exit() is invoked.

        Usually this would either start polling some API or implement the
        server side of a push based protocol, using a framework such as asyncio.

        Whenever a new message is available, just call post_message() with
        the new message.

        Returns:
            None
        """
        raise NotImplementedError

    def post_message(self, msg):
        """Post a message as an output for this source

        post_message() can be called from within run() to post messages to
        syslog-ng, as the output of the source driver being implemented.

        Flow control is implemented in two ways:
          - blocking mode: post_message() would block if flow control kicks
            in (e.g.  there's no room in our output window).  This is simple
            but your main event loop will be suspended until syslog-ng is
            able to accept messages again.

          - non-blocking mode: if suspend() and wakeup() methods are
            implemented in your class, post_message() would not block,
            rather invoke suspend() to indicate that no further messages
            is to be produced and wakeup() when it is possible to produce
            messages again.  This mode allows the continued execution of the
            async main loop even while syslog-ng is unable to accept
            messages.

        Arguments:
            msg: LogMessage
                the log message to be posted
        """
        super().post_message(msg)

    def request_exit(self):
        """Request the main loop to exit

        This function is invoked by syslog-ng from its main thread when this
        source should stop executing.  The expected result of this function
        is that its run() method returns to its caller.
        """
        raise NotImplementedError


class LogFetcher(LogFetcher):
    """Base class for building synchronous (blocking) source driver

    A LogFetcher is a base class that allows the implementation of a
    syslog-ng source driver.  This class starts a dedicated thread and
    regularly executes the fetch() method in that thread to fetch the next
    message to be emitted as the output of the log source.

    It easily maps to any source mechanism that you need to poll regularly
    for new messages (e.g.  HTTP based log access APIs).
    """
    ERROR = FETCH_ERROR = LogFetcherResult.ERROR
    NOT_CONNECTED = FETCH_NOT_CONNECTED = LogFetcherResult.NOT_CONNECTED
    SUCCESS = FETCH_SUCCESS = LogFetcherResult.SUCCESS
    TRY_AGAIN = FETCH_TRY_AGAIN = LogFetcherResult.TRY_AGAIN
    NO_DATA = FETCH_NO_DATA = LogFetcherResult.NO_DATA

    flags = {}
    parse_options = None

    def __init__(self):
        """Constructs this LogFetcher instance
        """
        super().__init__()

    def init(self, options):
        """Initialize this LogFetcher instance

        Returns:
            bool: True to indicate success
        """
        return True

    def deinit(self):
        """Deinitialize this LogFetcher instance
        """

    def open(self):
        """Open the connection to the target service

        This function is called by syslog-ng to open a connection to the
        target service.  It should return True to indicate success or False
        otherwise.  If the connection to the target service fails, syslog-ng
        will try to reconnect until it succeeds, retrying every
        time_reopen() seconds.
        """
        return True

    def close(self):
        """Close the connection to the target service"""

    def fetch(self):
        """Function to fetch the next message

        This function is executed from the thread associated with LogFetcher
        every time syslog-ng is able to consume a new message.

        Returns:
            (result, message): (LogFetcherResult, LogMessage)
                the function needs to return a 2-tuple, which comprises of a
                status code and a message instance.  Possible status codes
                are defined as members in the LogFetcher class or as values
                in the LogFetchrResult enum.
        """
        raise NotImplementedError

    def request_exit(self):
        """Request fetch() to exit

        This function is invoked by syslog-ng from its main thread when this
        source should stop executing.
        The method is optional, use it when fetch() has blocking operations and
        it has to be interrupted when shutting down syslog-ng.
        """


class InstantAckTracker(InstantAckTracker):
    def __init__(self, ack_callback):
        super().__init__(ack_callback=ack_callback)


class ConsecutiveAckTracker(ConsecutiveAckTracker):
    def __init__(self, ack_callback):
        super().__init__(ack_callback=ack_callback)


class BatchedAckTracker(BatchedAckTracker):
    def __init__(self, timeout, batch_size, batched_ack_callback):
        super().__init__(timeout=timeout, batch_size=batch_size, batched_ack_callback=batched_ack_callback)



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
#
# This module contains a stub for all functions exported by syslog-ng, so when
# the module is imported from within syslog-ng it uses the exported functions,
# in case we are running outside of syslog-ng we are using the stubs that is
# enough for testing.
#
from __future__ import print_function, absolute_import


def get_nv_registry():
    return ('0', '1', '2', '3', '4', "DATE", "HOST", "MSGHDR", "PROGRAM", "PID", "MSG", '.unix.uid', '.unix.gid')


def get_debugger_commands():
    return (
        "help",
        "continue",
        "print",
        "display",
        "drop",
        "quit"
    )


def get_template_functions():
    return (
        "python",
        "graphite-output",
        "uuid",
        "hash",
        "sha1",
        "sha256",
        "sha512",
        "md4",
        "md5",
        "format-json",
        "grep",
        "if",
        "or",
        "echo",
        "length",
        "substr",
        "strip",
        "sanitize",
        "lowercase",
        "uppercase",
        "replace-delimiter",
        "padding",
        "+",
        "-",
        "*",
        "/",
        "%",
        "ipv4-to-int",
        "indent-multi-line",
        "context-length",
        "env",
        "template"
    )


def get_value_pairs_scopes():
    return (
        "nv-pairs",
        "dot-nv-pairs",
        "all-nv-pairs",
        "rfc3164",
        "core",
        "base",
        "rfc5424",
        "syslog-proto",
        "all-macros",
        "selected-macros",
        "sdata",
        "everything",
    )


# override implementations from the module supplied by the C implementation.
try:
    # pylint: disable=import-error,wildcard-import,wrong-import-position
    from _syslogngdbg import *
except ImportError:
    pass



#############################################################################
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=unused-import,invalid-name

import logging

try:
    from _syslogng import Logger

    FAKE_LOGGER = False
except ImportError:
    import warnings
    warnings.warn("You have imported the syslogng package outside of syslog-ng, "
                  "thus some of the functionality is not available. "
                  "Defining fake classes for those exported by the underlying syslog-ng code")

    FAKE_LOGGER = True

    def Logger():
        return logging.getLogger()
    logging.basicConfig()


class SyslogNGInternalLogHandler(logging.Handler):
    syslogng_logger = Logger()

    def emit(self, record):

        msg = self.format(record)
        if record.levelno >= logging.ERROR:
            self.syslogng_logger.error(msg)
        elif record.levelno >= logging.WARNING:
            self.syslogng_logger.warning(msg)
        elif record.levelno >= logging.INFO:
            self.syslogng_logger.info(msg)
        elif record.levelno >= logging.DEBUG:
            self.syslogng_logger.debug(msg)
        else:
            self.syslogng_logger.trace(msg)


def setup_logging():
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    sh = SyslogNGInternalLogHandler()
    formatter = logging.Formatter('python-%(name)s: %(message)s')
    sh.setFormatter(formatter)
    logger.addHandler(sh)


if not FAKE_LOGGER:
    setup_logging()



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from .completerlang import CompleterLang
from .templatelexer import TemplateLexer


class TemplateLang(CompleterLang):

    tokens = [
        "LITERAL", "MACRO", "TEMPLATE_FUNC"
    ]

    @staticmethod
    def p_template_string(p):
        '''template_string : template_elems'''

    @staticmethod
    def p_template_elems(p):
        '''template_elems : template_elem template_elems
                          |'''

    @staticmethod
    def p_template_elem(p):
        '''template_elem : LITERAL
                         | MACRO
                         | TEMPLATE_FUNC
        '''

    def _construct_lexer(self):
        return TemplateLexer()



#############################################################################
# Copyright (c) 2020 One Identity
# Copyright (c) 2020 Laszlo Budai <laszlo.budai@outlook.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=import-error

from __future__ import absolute_import, print_function

import sys

from editline import _editline
from editline.editline import EditLine
from editline import lineeditor

from .debuggercli import DebuggerCLI


class EditlineCompleteHook():
    def __init__(self, completer, editl):
        self._completer = completer
        self._last_contents = (None, None)
        self._last_completions = []
        self.editl = editl

    def complete(self, text):
        return self._get_completions(self.editl.get_line_buffer(), text)

    def _get_completions(self, entire_text, text):
        if self._last_contents == (entire_text, text):
            return self._last_completions
        self._last_completions = self._completer.complete(entire_text, text)
        self._last_contents = (entire_text, text)
        return self._last_completions


class MyEditLineCompleter(lineeditor.Completer):
    def __init__(self, subeditor, completer, namespace=None):
        super().__init__(subeditor, namespace)
        self._default_display_matches = self.subeditor.display_matches
        self.subeditor.display_matches = self.display_matches
        self.completer = completer
        self.subeditor.completer = self.complete

    def complete(self, text):
        # pylint: disable=attribute-defined-outside-init
        self.matches = self.completer.complete(text)
        return self.matches

    def display_matches(self, matches):
        # pylint: disable=protected-access
        self.subeditor._display_matches(self.matches)


__setup_performed__ = False


def setup_editline():
    # pylint: disable=global-statement
    global __setup_performed__

    if __setup_performed__:
        return

    debuggercli = DebuggerCLI()

    editline_system = _editline.get_global_instance()
    if editline_system is None:
        sys_el = EditLine("DebuggerCLI", sys.stdin, sys.stdout, sys.stderr)
        _editline.set_global_instance(sys_el)
        completer = EditlineCompleteHook(debuggercli.get_root_completer(), sys_el)
        sys_line_ed = MyEditLineCompleter(sys_el, completer=completer)
        lineeditor.global_line_editor(sys_line_ed)
        sys_el.completer = sys_line_ed.complete

    __setup_performed__ = True



#############################################################################
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
# Copyright (c) 2015-2016 Balabit
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
# pylint: disable=unused-import,function-redefined

try:
    from _syslogng import Persist

except ImportError:
    import warnings
    warnings.warn("You have imported the syslogng package outside of syslog-ng, "
                  "thus some of the functionality is not available. "
                  "Defining fake classes for those exported by the underlying syslog-ng code")

    # fake Persist class
    class Persist(dict):
        def __init__(self, persist_name):
            self.persist_name = persist_name


class Persist(Persist):
    def __init__(self, persist_name, defaults=None):
        super().__init__(persist_name)

        if defaults:
            for key, value in defaults.items():
                if key not in self:
                    self[key] = value



#############################################################################
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import annotations

from .compressable_file_buffer import CompressableFileBuffer

try:
    from botocore.exceptions import ClientError, EndpointConnectionError
except ImportError:
    pass

from concurrent.futures import Future, ThreadPoolExecutor
from logging import Logger
from pathlib import Path
from threading import Event, Lock
from time import monotonic
from typing import Any, Dict, Optional, List, Set
from uuid import uuid4

import json


class PersistLoadError(Exception):
    pass


class ConstructorError(Exception):
    pass


class AlreadyFinishedError(Exception):
    pass


class S3Chunk:
    MAX_PART_NUMBER = 9999

    def __init__(
        self,
        part_number: Optional[int] = None,
        compress: bool = False,
        compresslevel: Optional[int] = None,
        working_dir: Optional[Path] = None,
        path: Optional[Path] = None,
        prev_chunk: Optional[S3Chunk] = None,
    ) -> None:
        if prev_chunk is not None:
            path = Path(prev_chunk.buffer.path.parent, str(uuid4()))
            self.__buffer = prev_chunk.buffer.create_next(path)
            self.__part_number = prev_chunk.part_number + 1
            return

        if path is not None:
            assert part_number is not None
            self.__buffer = CompressableFileBuffer.load_finished(path=path)
            self.__part_number = part_number
            return

        assert working_dir is not None
        path = Path(working_dir, str(uuid4()))

        assert compress is not None and compresslevel is not None and part_number is not None
        self.__buffer = CompressableFileBuffer.create_initial(path=path, compress=compress, compresslevel=compresslevel)
        self.__part_number = part_number

    @staticmethod
    def create_initial(working_dir: Path, compress: bool, compresslevel: int) -> S3Chunk:
        return S3Chunk(working_dir=working_dir, part_number=1, compress=compress, compresslevel=compresslevel)

    @staticmethod
    def load_finished(path: Path, part_number: int) -> S3Chunk:
        return S3Chunk(path=path, part_number=part_number)

    def create_next(self) -> S3Chunk:
        return S3Chunk(prev_chunk=self)

    @property
    def buffer(self) -> CompressableFileBuffer:
        return self.__buffer

    @property
    def part_number(self) -> int:
        return self.__part_number


class S3ObjectPersist:
    """Thread safe persist class for the S3Object class."""

    def __init__(
        self,
        working_dir: Optional[Path] = None,
        path: Optional[Path] = None,
        persist_name: Optional[str] = None,
        bucket: Optional[str] = None,
        target_key: Optional[str] = None,
        timestamp: Optional[str] = None,
        compress: Optional[bool] = None,
        compresslevel: Optional[int] = None,
        chunk_size: Optional[int] = None,
        storage_class: Optional[str] = None,
        canned_acl: Optional[str] = None,
    ):
        if path is not None:
            self.__path = path
        elif working_dir is not None:
            self.__path = Path(working_dir, f"{str(uuid4())}.json")
            assert (
                persist_name is not None
                and bucket is not None
                and target_key is not None
                and timestamp is not None
                and compress is not None
                and compresslevel is not None
                and chunk_size is not None
                and storage_class is not None
                and canned_acl is not None
            )
        else:
            assert False

        self.__path.resolve().parent.mkdir(parents=True, exist_ok=True)
        self.__path.touch()

        self.__lock = Lock()

        cache: Dict[str, Any] = dict()
        with self.__path.open("r", encoding="utf-8") as f:
            try:
                cache = json.load(f)
            except json.decoder.JSONDecodeError as e:
                if path:
                    raise PersistLoadError from e

        if path:
            for field in {
                "persist-name",
                "bucket",
                "target-key",
                "timestamp",
                "index",
                "compress",
                "compresslevel",
                "chunk-size",
                "storage-class",
                "finished",
                "upload-id",
                "uploaded-parts",
                "pending-parts",
            }:
                try:
                    cache[field]
                except KeyError as e:
                    raise PersistLoadError from e

        self.__persist_name: str = cache.get("persist-name", persist_name)
        self.__bucket: str = cache.get("bucket", bucket)
        self.__target_key: str = cache.get("target-key", target_key)
        self.__timestamp: str = cache.get("timestamp", timestamp)
        self.__index: int = cache.get("index", -1)
        self.__compress: bool = cache.get("compress", compress)
        self.__compresslevel: bool = cache.get("compresslevel", compresslevel)
        self.__chunk_size: bool = cache.get("chunk-size", chunk_size)
        self.__storage_class: str = cache.get("storage-class", storage_class)
        self.__canned_acl: str = cache.get("canned-acl", canned_acl)
        self.__finished: bool = cache.get("finished", False)
        self.__upload_id: str = cache.get("upload-id", "")
        self.__uploaded_parts: List[Dict[str, Any]] = cache.get("uploaded-parts", [])
        self.__pending_parts: Dict[str, Any] = cache.get("pending-parts", dict())

        self.__flush()

    @staticmethod
    def load(path: Path) -> S3ObjectPersist:
        return S3ObjectPersist(path=path)

    def __flush(self) -> None:
        cache = {
            "persist-name": self.__persist_name,
            "bucket": self.__bucket,
            "target-key": self.__target_key,
            "timestamp": self.__timestamp,
            "index": self.__index,
            "compress": self.__compress,
            "compresslevel": self.__compresslevel,
            "storage-class": self.__storage_class,
            "canned-acl": self.__canned_acl,
            "chunk-size": self.__chunk_size,
            "finished": self.__finished,
            "upload-id": self.__upload_id,
            "uploaded-parts": self.__uploaded_parts,
            "pending-parts": self.__pending_parts,
        }

        with self.__path.open("w", encoding="utf-8") as f:
            try:
                json.dump(cache, f)
                f.flush()
            except TypeError:
                assert False

    def unlink(self):
        self.__path.unlink()

    @property
    def path(self) -> Path:
        return self.__path

    @property
    def persist_name(self) -> str:
        return self.__persist_name

    @property
    def bucket(self) -> str:
        return self.__bucket

    @property
    def target_key(self) -> str:
        return self.__target_key

    @property
    def timestamp(self) -> str:
        return self.__timestamp

    @property
    def index(self) -> int:
        with self.__lock:
            return self.__index

    @index.setter
    def index(self, i: int) -> None:
        with self.__lock:
            self.__index = i
            self.__flush()

    @property
    def compress(self) -> bool:
        return self.__compress

    @property
    def compresslevel(self) -> bool:
        return self.__compresslevel

    @property
    def chunk_size(self) -> bool:
        return self.__chunk_size

    @property
    def storage_class(self) -> str:
        return self.__storage_class

    @property
    def canned_acl(self) -> str:
        return self.__canned_acl

    @property
    def finished(self) -> bool:
        with self.__lock:
            return self.__finished

    @finished.setter
    def finished(self, f: bool) -> None:
        with self.__lock:
            self.__finished = f
            self.__flush()

    @property
    def upload_id(self) -> str:
        with self.__lock:
            return self.__upload_id

    @upload_id.setter
    def upload_id(self, u: str) -> None:
        with self.__lock:
            self.__upload_id = u
            self.__flush()

    @property
    def pending_parts(self) -> Dict[str, Any]:
        """
        A shallow copy of the pending parts.
        """
        with self.__lock:
            return self.__pending_parts.copy()

    def add_pending_part(self, path: Path, part_number: int) -> None:
        with self.__lock:
            self.__pending_parts[str(path)] = {"number": part_number}
            self.__flush()

    def cancel_pending_part(self, path: Path) -> None:
        with self.__lock:
            del self.__pending_parts[str(path)]
            self.__flush()

    @property
    def uploaded_parts(self) -> List[Dict[str, Any]]:
        """
        A sorted shallow copy of the uploaded parts.
        Can be used directly at the finish_multipart_upload() call.
        """
        with self.__lock:
            self.__uploaded_parts = sorted(self.__uploaded_parts, key=lambda x: x["PartNumber"])
            return self.__uploaded_parts.copy()

    def part_upload_finished(self, path: Path, etag: str, part_number: int) -> None:
        with self.__lock:
            self.__uploaded_parts.append({"ETag": etag, "PartNumber": part_number})
            del self.__pending_parts[str(path)]
            self.__flush()


class S3Object:
    MIN_CHUNK_SIZE_BYTES = 5 * 1024 * 1024

    def __init__(
        self,
        executor: ThreadPoolExecutor,
        client: Any,
        logger: Logger,
        exit_requested: Event,
        bucket: Optional[str] = None,
        target_key: Optional[str] = None,
        timestamp: Optional[str] = None,
        target_index: Optional[int] = None,
        compress: Optional[bool] = None,
        chunk_size: Optional[int] = None,
        compresslevel: Optional[int] = None,
        storage_class: Optional[str] = None,
        canned_acl: Optional[str] = None,
        persist: Optional[S3ObjectPersist] = None,
        persist_name: Optional[str] = None,
        working_dir: Optional[Path] = None,
    ) -> None:
        self.__executor = executor
        self.__client = client
        self.__logger = logger
        self.__exit_requested = exit_requested

        self.__size: int = 0
        self.__modified_at: float = monotonic()

        # Locked variables
        self.__lock = Lock()
        self.__current_chunk: Optional[S3Chunk] = None
        self.__upload_futures: List[Future] = []
        self.__multipart_completed: bool = False
        self.__multipart_complete_future: Optional[Future] = None

        if persist is None:
            assert (
                persist_name is not None
                and working_dir is not None
                and bucket is not None
                and target_key is not None
                and timestamp is not None
                and target_index is not None
                and compress is not None
                and chunk_size is not None
                and compresslevel is not None
                and storage_class is not None
                and canned_acl is not None
            )
            self.__persist = S3ObjectPersist(
                working_dir=working_dir,
                persist_name=persist_name,
                bucket=bucket,
                target_key=target_key,
                timestamp=timestamp,
                compress=compress,
                compresslevel=compresslevel,
                chunk_size=chunk_size,
                storage_class=storage_class,
                canned_acl=canned_acl,
            )
            self.__persist.index = self.__get_next_index(target_index)

            if self.index == -1:
                self.__persist.unlink()
                raise ConstructorError(f"Failed to get next ID for: {self.bucket}/{self.target_key}")

            self.__current_chunk = S3Chunk.create_initial(
                working_dir=working_dir,
                compress=compress,
                compresslevel=compresslevel,
            )
            self.__persist.add_pending_part(self.__current_chunk.buffer.path, self.__current_chunk.part_number)
        else:
            self.__persist = persist
            if not self.__persist.finished and self.__persist.compress:
                self.__logger.error(
                    f"Loaded an unfinished compressed object: {self.__persist.bucket}/{self.key}. "
                    "It is probably in an unusable state. syslog-ng will still upload this file, "
                    "but expect problems, while decompressing the file."
                )

            if self.bucket == "" or self.index == -1 or self.target_key == "":
                raise ConstructorError(f"Failed to load persist values for: {self.bucket}/{self.target_key}")

            for path, data in self.__persist.pending_parts.items():
                chunk = S3Chunk.load_finished(path=Path(path), part_number=data["number"])
                self.__upload_chunk(chunk)

            self.__persist.finished = True
            self.__complete_multipart()

    @staticmethod
    def create_initial(
        working_dir: Path,
        bucket: str,
        target_key: str,
        timestamp: str,
        compress: bool,
        compresslevel: int,
        storage_class: str,
        canned_acl: str,
        persist_name: str,
        chunk_size: int,
        executor: ThreadPoolExecutor,
        client: Any,
        logger: Logger,
        exit_requested: Event,
    ) -> S3Object:
        return S3Object(
            working_dir=working_dir,
            executor=executor,
            client=client,
            logger=logger,
            bucket=bucket,
            target_key=target_key,
            timestamp=timestamp,
            target_index=0,
            compress=compress,
            compresslevel=compresslevel,
            storage_class=storage_class,
            canned_acl=canned_acl,
            chunk_size=chunk_size,
            persist_name=persist_name,
            exit_requested=exit_requested,
        )

    @staticmethod
    def load_finished(
        persist: S3ObjectPersist,
        executor: ThreadPoolExecutor,
        client: Any,
        logger: Logger,
        exit_requested: Event,
    ) -> S3Object:
        return S3Object(persist=persist, executor=executor, client=client, logger=logger, exit_requested=exit_requested)

    def create_next(self) -> S3Object:
        return S3Object(
            working_dir=self.__persist.path.parent,
            executor=self.__executor,
            client=self.__client,
            logger=self.__logger,
            bucket=self.bucket,
            target_key=self.target_key,
            timestamp=self.timestamp,
            target_index=self.index + 1,
            compress=self.__persist.compress,
            compresslevel=self.__persist.compresslevel,
            storage_class=self.__persist.storage_class,
            canned_acl=self.__persist.canned_acl,
            chunk_size=self.__persist.chunk_size,
            persist_name=self.__persist.persist_name,
            exit_requested=self.__exit_requested,
        )

    @staticmethod
    def format_key(target_key: str, timestamp: str, index: int, compression: bool) -> str:
        key = target_key
        key += f"-{timestamp}" if timestamp != "" else ""
        key += f"-{str(index)}" if index > 0 else ""
        key += ".gz" if compression else ""
        return key

    @property
    def bucket(self) -> str:
        return self.__persist.bucket

    @property
    def target_key(self) -> str:
        return self.__persist.target_key

    @property
    def key(self) -> str:
        return self.format_key(self.target_key, self.timestamp, self.index, self.__persist.compress)

    @property
    def timestamp(self) -> str:
        return self.__persist.timestamp

    @property
    def index(self) -> int:
        return self.__persist.index

    @property
    def finished(self) -> bool:
        return self.__persist.finished

    @property
    def size(self) -> int:
        return self.__size

    @property
    def modified_at(self) -> float:
        return self.__modified_at

    @property
    def uploaded_parts(self) -> List[Dict[str, Any]]:
        """
        Waits for part uploads to finish, then returns a shallow copy of the uploaded parts.
        Can be used directly at the finish_multipart_upload() function.
        """
        while True:
            with self.__lock:
                try:
                    future = self.__upload_futures.pop()
                except IndexError:
                    break
            future.result()
        return self.__persist.uploaded_parts

    def __ensure_multipart_upload_started(self) -> bool:
        if self.__persist.upload_id != "":
            return True

        try:
            if self.__persist.canned_acl != "":
                response = self.__client.create_multipart_upload(
                    Bucket=self.bucket,
                    Key=self.key,
                    StorageClass=self.__persist.storage_class,
                    ACL=self.__persist.canned_acl,
                )
            else:
                response = self.__client.create_multipart_upload(
                    Bucket=self.bucket,
                    Key=self.key,
                    StorageClass=self.__persist.storage_class,
                )
        except (ClientError, EndpointConnectionError) as e:
            self.__logger.error(f"Failed to create multipart upload: {self.bucket}/{self.key} => {e}")
            return False

        try:
            self.__persist.upload_id = response["UploadId"]
        except KeyError:
            self.__logger.error(
                f"Failed to create multipart upload: {self.bucket}/{self.key} => Unexpected response: {response}"
            )
            return False

        return True

    def __upload_chunk_cb(self, chunk: S3Chunk, is_retry: bool) -> None:
        if not self.__ensure_multipart_upload_started():
            self.__upload_chunk(chunk, is_retry=True)
            return

        try:
            response = self.__client.upload_part(
                Bucket=self.bucket,
                Key=self.key,
                UploadId=self.__persist.upload_id,
                PartNumber=chunk.part_number,
                Body=chunk.buffer.getvalue(),
            )
        except EndpointConnectionError as e:
            self.__logger.error(f"Failed to upload part: {self.bucket}/{self.key} ({chunk.part_number}) => {e}")
            self.__upload_chunk(chunk, is_retry=True)
            return
        except ClientError as e:
            self.__logger.error(
                f"Critical failure when trying to upload part, finishing S3 object: {self.bucket}/{self.key} => {e}"
            )
            self.__persist.cancel_pending_part(chunk.buffer.path)
            self.finish()
            return

        try:
            etag = response["ETag"]
        except KeyError:
            self.__logger.error(
                f"Failed to upload part: {self.bucket}/{self.key} ({chunk.part_number}) => Unexpected response: {response}"
            )
            self.__upload_chunk(chunk, is_retry=True)
            return

        if is_retry:
            self.__logger.info(
                f"Successfully uploaded part after retry: {self.bucket}/{self.key} ({chunk.part_number})"
            )

        self.__persist.part_upload_finished(chunk.buffer.path, etag, chunk.part_number)

        chunk.buffer.unlink()

    def __upload_chunk(self, chunk: S3Chunk, is_retry: bool = False) -> None:
        if is_retry and self.__exit_requested.is_set():
            self.__logger.debug(
                f"Exit requested, skipping chunk upload: {self.bucket}/{self.key} => "
                f"{chunk.part_number} - {chunk.buffer.path}"
            )
            return

        if self.finished and chunk.buffer.is_empty():
            self.__persist.cancel_pending_part(chunk.buffer.path)
            chunk.buffer.unlink()
            return

        future = self.__executor.submit(self.__upload_chunk_cb, chunk, is_retry)
        with self.__lock:
            self.__upload_futures.append(future)

    def write(self, data: bytes) -> None:
        """Raises OSError on write failure. Cannot be called from multiple threads."""
        with self.__lock:
            chunk = self.__current_chunk
        if chunk is None:
            raise AlreadyFinishedError()

        old_size = chunk.buffer.tell()
        chunk.buffer.write(data)
        new_size = chunk.buffer.tell()
        self.__size += new_size - old_size

        self.__modified_at = monotonic()

        if new_size > self.__persist.chunk_size:
            self.__lock.acquire()

            if self.__current_chunk is None:
                # finish() was called while we were writing the buffer, this can only happen in part upload failure.
                self.__lock.release()
                raise AlreadyFinishedError()

            if self.__current_chunk.part_number == S3Chunk.MAX_PART_NUMBER:
                self.__lock.release()
                self.finish()
                return

            self.__current_chunk = chunk.create_next()
            self.__persist.add_pending_part(self.__current_chunk.buffer.path, self.__current_chunk.part_number)

            self.__lock.release()

            self.__upload_chunk(chunk)

    def __abort_multipart(self) -> bool:
        assert self.__persist.upload_id

        try:
            self.__client.abort_multipart_upload(
                Bucket=self.bucket,
                Key=self.key,
                UploadId=self.__persist.upload_id,
            )
        except EndpointConnectionError as e:
            self.__logger.error(f"Failed to abort multipart upload: {self.bucket}/{self.key} => {e}")
            return False
        except ClientError as e:
            self.__logger.error(f"Failed to abort multipart upload: {self.bucket}/{self.key} => {e}")
            if e.response["Error"]["Code"] == "NoSuchUpload":
                return True
            return False

        return True

    def __complete_multipart_cb(self, is_retry: bool) -> None:
        uploaded_parts = self.uploaded_parts
        pending_parts = self.__persist.pending_parts

        if len(pending_parts) > 0:
            self.__logger.error(
                f"Failed to complete multipart upload: {self.bucket}/{self.key} => Part uploads still pending"
            )
            self.__complete_multipart(is_retry=True)
            return

        if len(uploaded_parts) == 0:
            assert self.__persist.upload_id == ""
            self.__persist.unlink()
            with self.__lock:
                self.__multipart_completed = True
            return

        assert self.__persist.upload_id

        try:
            self.__client.complete_multipart_upload(
                Bucket=self.bucket,
                Key=self.key,
                MultipartUpload={"Parts": uploaded_parts},
                UploadId=self.__persist.upload_id,
            )
        except EndpointConnectionError as e:
            self.__logger.error(f"Failed to complete multipart upload: {self.bucket}/{self.key} => {e}")
            self.__complete_multipart(is_retry=True)
            return
        except ClientError as e:
            if e.response["Error"]["Code"] in {"EntityTooSmall", "InvalidPart", "InvalidPartOrder", "NoSuchUpload"}:
                self.__logger.error(
                    f"Critical failure when trying to complete multipart upload, dropping S3 object: {self.bucket}/{self.key} => {e}"
                )

                if not self.__abort_multipart():
                    self.__complete_multipart(is_retry=True)
                    return

                self.__persist.unlink()
                with self.__lock:
                    self.__multipart_completed = True
                return

            self.__logger.error(f"Failed to complete multipart upload: {self.bucket}/{self.key} => {e}")
            self.__complete_multipart(is_retry=True)
            return

        if is_retry:
            self.__logger.info(f"Successfully completed multipart upload after retry: {self.bucket}/{self.key}")

        self.__persist.unlink()
        with self.__lock:
            self.__multipart_completed = True

    def __complete_multipart(self, is_retry: bool = False) -> None:
        if is_retry and self.__exit_requested.is_set():
            self.__logger.debug(f"Exit requested, skipping multipart upload completion: {self.bucket}/{self.key}")
            with self.__lock:
                self.__multipart_complete_future = None
            return

        with self.__lock:
            self.__multipart_complete_future = self.__executor.submit(self.__complete_multipart_cb, is_retry)

    def finish(self) -> None:
        if self.finished:
            return

        chunk: Optional[S3Chunk] = None
        with self.__lock:
            if self.__current_chunk:
                chunk = self.__current_chunk
                self.__current_chunk.buffer.finish()
                self.__current_chunk = None

        if chunk is not None:
            self.__upload_chunk(chunk)

        self.__persist.finished = True
        self.__complete_multipart()

    def wait_for_upload_to_complete(self) -> None:
        assert self.finished

        is_retry = False
        while True:
            with self.__lock:
                if self.__multipart_complete_future is None:
                    break
                if self.__multipart_completed:
                    break
                if is_retry and self.__exit_requested.is_set():
                    break
                future = self.__multipart_complete_future
            future.result()
            is_retry = True

    def __list_pending_multiparts_in_s3(self) -> Optional[Set[str]]:
        uploads: Set[str] = set()
        key_marker: str = ""
        upload_id_marker: str = ""

        while True:
            try:
                response: Dict[str, Any] = self.__client.list_multipart_uploads(
                    Bucket=self.bucket,
                    Prefix=self.target_key,
                    KeyMarker=key_marker,
                    UploadIdMarker=upload_id_marker,
                )
            except (ClientError, EndpointConnectionError) as e:
                self.__logger.error(f"Failed to list multipart uploads: {self.bucket}/{self.key} => {e}")
                return None

            try:
                for obj in response.get("Uploads", []):
                    uploads.add(obj["Key"])
                if not response["IsTruncated"]:
                    break
                key_marker = response["NextKeyMarker"]
                upload_id_marker = response["NextUploadIdMarker"]
            except KeyError:
                self.__logger.error(
                    f"Failed to list multipart uploads: {self.bucket}/{self.key} => Unexpected response: {response}"
                )
                return None

        return uploads

    def __list_existing_objects_in_s3(self) -> Optional[Set[str]]:
        objects: Set[str] = set()
        marker: str = ""

        while True:
            try:
                response: Dict[str, Any] = self.__client.list_objects(
                    Bucket=self.bucket,
                    Prefix=self.target_key,
                    Marker=marker,
                )
            except (ClientError, EndpointConnectionError) as e:
                self.__logger.error(f"Failed to list objects: {self.bucket}/{self.key} => {e}")
                return None

            try:
                for obj in response.get("Contents", []):
                    objects.add(obj["Key"])
                if not response["IsTruncated"]:
                    break
                marker = response["Contents"][-1]["Key"]
            except KeyError:
                self.__logger.error(
                    f"Failed to list objects: {self.bucket}/{self.key} => Unexpected response: {response}"
                )
                return None

        return objects

    def __get_next_index(self, target_index: int) -> int:
        pending_multiparts = self.__list_pending_multiparts_in_s3()
        if pending_multiparts is None:
            return -1

        existing_objects = self.__list_existing_objects_in_s3()
        if existing_objects is None:
            return -1

        objects: Set[str] = pending_multiparts | existing_objects

        i = target_index
        while True:
            uncompressed_key = S3Object.format_key(self.target_key, self.timestamp, i, False)
            compressed_key = S3Object.format_key(self.target_key, self.timestamp, i, True)
            if uncompressed_key not in objects and compressed_key not in objects:
                return i
            i += 1



#############################################################################
# Copyright (c) 2022 novaSOC
# Copyright (c) 2022 Balazs Scheidler <bazsi77@gmail.com>
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
#
# This was the original license on the code taken from:
#     https://github.com/danelder/syslogng-hypr
#
"""
Copyright (c) 2022 novaSOC

Use of this source code is governed by an MIT-style
license that can be found in the LICENSE file or at
https://opensource.org/licenses/MIT.

Original development by Dan Elder (delder@novacoast.com)
Syslog-ng python fetcher for Hypr API (https://apidocs.hypr.com)

Additional documentation available at:
https://www.syslog-ng.com/technical-documents/doc/syslog-ng-open-source-edition/3.36/administration-guide/26#TOPIC-1768583
"""

import time
import json
import logging
import base64
import re
from datetime import datetime, timezone

try:
    import requests
except ImportError:
    requests = None

import syslogng


class HyprAuditSource(syslogng.LogFetcher):
    """
    class for python syslog-ng log fetcher
    """

    # mandatory settings
    url = None
    bearer_token = None
    rp_app_id = None
    persist_name = None

    # optional settings
    page_size = 100
    initial_hours = 4

    # state
    start_time = 0
    end_time = 0

    #
    logger = logging.getLogger('hypr')
    persist = None

    def init(self, options):
        """
        Initialize Hypr driver
        (optional for Python LogFetcher)
        """

        if requests is None:
            self.logger.error("Unable to start HyprAuditSource class, the required Python dependency of `requests` is missing")
            return False

        # Ensure rp_app_id parameter is defined
        if "rp_app_id" in options:
            self.rp_app_id = options["rp_app_id"]
        else:
            self.logger.error("Missing rp_app_id configuration option for Hypr driver")
            return False

        # Initialize rp_app_id specific logger
        self.logger = logging.getLogger('hypr.' + self.rp_app_id)
        if "log_level" in options:
            self.logger.setLevel(options["log_level"])

        self.logger.info("Starting Hypr API fetch driver for %s", self.rp_app_id)

        # Initialize empty array of log messages
        self.logs = []

        # Ensure url parameter is defined
        if "url" in options:
            self.url = options["url"]
            self.logger.debug("Initializing Hypr %s syslog-ng driver against URL %s", self.rp_app_id, self.url)
        else:
            self.logger.error("Missing url configuration option for %s", self.rp_app_id)
            return False

        # Ensure bearer_token parameter is defined
        if "bearer_token" in options:
            self.token = options["bearer_token"]
            self.logger.debug("Initializing Hypr syslog-ng driver with bearer_token %s for %s",
                              self.token, self.rp_app_id)
        else:
            self.logger.error("Missing bearer_token configuration option for %s", self.rp_app_id)
            return False

        # Set page size if defined
        self.page_size = 100
        if "page_size" in options:
            self.page_size = options["page_size"]
            self.logger.debug("Initializing Hypr syslog-ng driver with pageSize %s for %s",
                              self.page_size, self.rp_app_id)

        # Default time to go back is 4 hours
        initial_hours = 4
        if "initial_hours" in options:
            initial_hours = options["initial_hours"]
            self.logger.debug("Initializing Hypr syslog-ng driver with initial_hours %i hours ago for %s",
                              initial_hours, self.rp_app_id)

        # Start with last search window end time or ignore
        ignore_persistence = False
        if "ignore_persistence" in options:
            ignore_persistence = options["ignore_persistence"]
            self.logger.debug("Initializing Hypr syslog-ng driver with %spersistence for %s",
                              "ignoring " if ignore_persistence else "", self.rp_app_id)

        # Convert initial_hours to milliseconds and subtract from current time
        self.start_time = int(time.time() * 1000) - (initial_hours * 3600000)

        # Setup persist_name with defined persist_name or use URL and rpAppId if none specified
        if self.persist_name is None:
            self.persist_name = "hypr-%s-%s" % (self.url, self.rp_app_id)

        # Initialize persistence
        self.logger.debug("Initializing Hypr syslog-ng driver with persist_name %s", self.persist_name)
        self.persist = syslogng.Persist(persist_name=self.persist_name, defaults={"last_read": self.start_time})

        # Initialize ack tracing
        self.ack_tracker = syslogng.ConsecutiveAckTracker(ack_callback=self.message_acked)

        # Convert persistence timestamp and reset if invalid data is in persistence
        try:
            last_run = datetime.utcfromtimestamp(int(self.persist["last_read"]) / 1000)
            self.logger.debug("Read %s from persistence as last run time", last_run)
        except (OverflowError, OSError):
            self.logger.error("Invalid last_read detected in persistence, resetting to %s hours ago and ignoring "
                              "persistence for %s", initial_hours, self.rp_app_id)
            ignore_persistence = True

        # Ignore persistence if configured
        if not ignore_persistence:
            # Start search at last fetch window end time
            self.start_time = int(self.persist["last_read"])

        self.logger.debug("Driver initialization complete, fetch window starts at %i (%s)",
                          self.start_time, datetime.utcfromtimestamp(self.start_time / 1000))

        return True

    def message_acked(self, acked_message_bookmark):
        self.persist["last_read"] = acked_message_bookmark

    def extract_nvpairs(self, msg, log, prefix, parent_name=""):
        for name, value in log.items():
            new_name = "%s.%s" % (parent_name, name) if parent_name else name
            value_type = type(value)

            if value_type == dict:
                self.extract_nvpairs(msg, value, prefix, new_name)
            elif ((value_type in (str, bytes, int, float, bool, datetime)) or
                  (value is None) or
                  (value_type == list and all(type(element) in (str, bytes) for element in value))):
                msg[prefix + new_name] = value
            else:
                msg[prefix + new_name] = str(value)

    def parse_log(self, log):
        """
        Parse an event into a syslog LogMessage
        (custom function for message parsing)
        """

        def simple_parse_log(log, rp_app_id):
            msg = syslogng.LogMessage("%s" % json.dumps(log))
            program = "Hypr-" + rp_app_id
            msg["PROGRAM"] = program.replace(" ", "-")

            return msg

        # Do not extract fields, if flags(no-parse) is set
        if not self.flags.get("parse", True):
            return simple_parse_log(log, self.rp_app_id)

        # Parse everything we can out of the message
        else:
            try:
                msg = syslogng.LogMessage(log["message"])
            except KeyError:
                self.logger.debug("'message' field is missing, dumping JSON data to ${MESSAGE} from %s", self.rp_app_id)
                return simple_parse_log(log, self.rp_app_id)

            # Try to get program/rpAppId from message
            if "rpAppId" in log:
                msg["PROGRAM"] = log['rpAppId']

            # Try to get timestamp information from message
            if "eventTimeInUTC" in log:
                try:
                    timestamp = datetime.fromtimestamp(int(log['eventTimeInUTC'] / 1000.0), tz=timezone.utc)
                    msg.set_timestamp(timestamp)

                except Exception as e_all:
                    self.logger.debug("Unable to convert %s to timestamp from %s : %s",
                                      log['eventTimeInUTC'], self.rp_app_id, e_all)

            self.extract_nvpairs(msg=msg, log=log, prefix=".hypr.")

        return msg

    def fetch(self):
        """
        Return a log message by pulling from the internal list or pulling from the Hypr API
        (mandatory function for Python LogFetcher)
        """

        # Retrieve log messages from memory if present
        if self.logs:
            log = self.logs.pop(0)

            if "eventTimeInUTC" not in log:
                self.logger.error("'eventTimeInUTC' field is missing, skipping message: %s" % str(log))
                return self.ERROR, None

            msg = self.parse_log(log)
            msg.set_bookmark(log["eventTimeInUTC"])
            return self.SUCCESS, msg

        # Get current time in milliseconds since epoch
        self.end_time = int(time.time() * 1000)

        # Retrieve log messages from Hypr API
        subscription_url = (self.url + "/cc/api/versioned/audit/search?" +
                            "rpAppId=" + self.rp_app_id +
                            "&startTSUTC=" + str(self.start_time) +
                            "&endTSUTC=" + str(self.end_time) +
                            "&pageSize=" + str(self.page_size) +
                            "&orderBy=eventTimeInUTC" +
                            "&sortDir=asc")

        headers = {"Content-Type": "application/application-json",
                   "Accept": "application/json",
                   "Authorization": "Bearer %s" % self.bearer_token}

        # Perform HTTP request
        response = requests.get(subscription_url, headers=headers)

        # Ingore 504 errors
        if response.status_code == 504:
            self.logger.info("Gateway Timeout from Hypr")
            return self.TRY_AGAIN, None

        # If the API call returns successfully, parse the retrieved json data
        if response.status_code == 200:

            try:
                # This JSON parsing is unfortunately necessary, as we get a batch of messages in one response.
                # There is no way to limit the level of the parsing, so we get each message as a python dict.
                result = response.json()
                total_records = result['metadata']['totalRecords']
                total_pages = result['metadata']['totalPages']
                current_page = result['metadata']['currentPage']

                # Set internal log buffer to all returned events
                self.logs = result['data']
                self.logger.debug("%i events available from Hypr API %s fetch", total_records, self.rp_app_id)
            except Exception as e_all:
                self.logger.error("%s - %s access failure : %s\n%s" % (self.url, self.rp_app_id, e_all, response.text))
                return self.ERROR, None

            # If there are more pages of events to process
            while current_page < total_pages:

                # increment page counter
                current_page = current_page + 1

                # Retrieve log messages from Hypr API
                subscription_url = (self.url + "/cc/api/versioned/audit/search?" +
                                    "rpAppId=" + self.rp_app_id +
                                    "&startTSUTC=" + str(self.start_time) +
                                    "&endTSUTC=" + str(self.end_time) +
                                    "&pageSize=" + str(self.page_size) +
                                    "&pageNumber=" + str(current_page))

                headers = {"Content-Type": "application/application-json",
                           "Accept": "application/json",
                           "Authorization": "Bearer %s" % self.bearer_token}

                # Perform HTTP request
                response = requests.get(subscription_url, headers=headers)

                # If we were successful, parse the json result
                if response.status_code == 200:
                    try:
                        result = response.json()
                    except Exception as e_all:
                        self.logger.error("%s - %s access failure : %s\n%s" % (self.url, self.rp_app_id, e_all, response.text))
                        return self.ERROR, None

                    # Add each event to our internal logs list
                    for entry in result['data']:
                        self.logs.append(entry)

                # If something went wrong with the query
                else:
                    self.logger.error("%s - %s access failure : %s\n%s" % (self.url, self.rp_app_id, response.status_code, response.text))
                    return self.ERROR, None

            # Set start time to end time
            self.start_time = self.end_time

            # If there are new logs
            if self.logs:
                return self.fetch()

            # If there aren't new logs
            self.logger.debug("No new Hypr events available")
            return self.NO_DATA, None

        # If the bearer token is invalid
        if response.status_code == 403:
            self.logger.error("Bearer token invalid for %s [%s]" % (self.rp_app_id, response.text))
            return self.ERROR, None

        # If the response code isn't 504 or 200 (or isn't even set)
        self.logger.error("%s - %s access failure : %s\n%s" % (self.url, self.rp_app_id, response.status_code, response.text))
        return self.ERROR, None

    def open(self):
        """
        Retrieve bearer token for Hypr API
        (optional for Python LogFetcher)
        """
        self.logger.info("Retreiving bearer token for Hypr API for %s", self.rp_app_id)
        try:
            self.bearer_token = base64.b64decode(self.token).decode("utf-8")
        except Exception as e_all:
            self.logger.error("Unable to decode bearer_token %s : %s", self.token, e_all)
            return False

        return True


class HyprError(Exception):
    pass


def _hypr_config_generator(args):
    if requests is None:
        raise HyprError("Unable to run confgen for hypr-audit-trail(), the required Python dependency `requests` is missing")
    logger = logging.getLogger("Hypr-Confgen")

    def sanitize(variable):
        """Filter out characters that would break syslog-ng configuration"""
        return variable.translate({ord(i): None for i in '\'\"\r'})

    # Capture environment variables for syslog-ng configuration
    url = sanitize(args.get('url', ""))
    bearer_token = sanitize(args.get('bearer_token', ""))
    page_size = args.get('page_size', "100")
    initial_hours = args.get('initial_hours', "4")
    fetch_no_data_delay = args.get('fetch_no_data_delay', "60")
    log_level = sanitize(args.get('log_level', "INFO"))
    application_skiplist = sanitize(args.get('application_skiplist',
                                    "'HYPRDefaultApplication' 'HYPRDefaultWorkstationApplication'")).replace(",", " ").split()
    ignore_persistence = args.get('ignore_persistence', 'no')
    persist_name = sanitize(args.get('persist_name', "hypr-%s" % url))
    flags = sanitize(args.get('flags', ''))

    # Log environment variables
    logger.debug("url : %s" % url)
    logger.debug("bearer_token : %s" % bearer_token)
    logger.debug("page_size : %s" % page_size)
    logger.debug("initial_hours : %s" % initial_hours)
    logger.debug("fetch_no_data_delay : %s" % fetch_no_data_delay)
    logger.debug("log_level : %s" % log_level)
    logger.debug("application_skiplist : %s" % application_skiplist)
    logger.debug("ignore_persistence: %s" % ignore_persistence)
    logger.debug("persist_name : %s" % persist_name)
    logger.debug("flags : %s" % flags)

    # Deobfuscate
    try:
        decoded_bearer_token = base64.b64decode(bearer_token).decode("utf-8")
    except Exception as e_all:
        raise ValueError("Unable to decode bearer_token %s : %s\n" % (bearer_token, e_all))

    # Start with empty list of applications
    applications = []

    # Get list of available applications from Hypr through API
    get_url = url + "/cc/api/application"
    headers = {
        "Content-Type": "application/application-json",
        "Authorization": "Bearer %s" % decoded_bearer_token
    }

    response = requests.get(get_url, headers=headers)

    # Check for valid response
    if response.status_code != 200:
        logger.debug("Hypr API fetch returned:%s" % response.text)
        logger.debug("Unable to retrieve applications from Hypr : HTTP %s" % response.status_code)
        logger.debug("Decoded bearer_token is %s" % decoded_bearer_token)
        logger.debug("Requested URL is %s" % get_url)
        raise HyprError("Unable to process hypr() source, HTTP error while retrieving application list: HTTP {} [{}]".format(response.status_code, response.text))

    # Loop through json response to pull all appIDs
    try:
        result = response.json()

        # Add appID to class list
        for entry in result:
            if entry['appID'] not in application_skiplist:
                applications.append(entry['appID'])
                logger.debug("Adding %s to monitored Hypr applications" % entry['appID'])

    except Exception as e_all:
        raise HyprError("Unable to retrieve list of applications from Hypr : %s" % e_all)

    # Setup source for each Hypr application
    sources = ""
    for application in applications:
        sources = sources + """
        python-fetcher(
            class("syslogng.modules.hypr.HyprAuditSource")
            options(
                "url" => "%s"
                "rp_app_id" => "%s"
                "bearer_token" => "%s"
                "page_size" => %s
                "initial_hours" => %s
                "ignore_persistence" => %s
                "log_level" => "%s"
            )
            flags(%s)
            persist-name("%s-%s")
            fetch-no-data-delay(%s)
        );
    """ % (url, application, bearer_token, page_size,
           initial_hours, ignore_persistence, log_level, flags,
           persist_name, application, fetch_no_data_delay)

    logger.debug("Final configuration is: %s" % sources)

    return sources


def register_hypr_config_generator():
    syslogng.register_config_generator(context="source", name="hypr-audit-trail", config_generator=_hypr_config_generator)



#############################################################################
# Copyright (c) 2022 Attila Szakacs <szakacs.attila96@gmail.com>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from kubernetes.client.rest import ApiException
from syslogng import Logger, LogParser
import kubernetes
import logging


class KubernetesAPIEnrichment(LogParser):
    logger = logging.getLogger(__name__)

    def init(self, options):
        self.__metadata = {}
        self.__prefix = options["prefix"]

        kubernetes.config.load_config()
        self.key_delimiter = options.get('key_delimiter', '.')

        self.__client_api = kubernetes.client.CoreV1Api()

        return True

    def add_prefix(self, string):
        return self.__prefix + string

    def __get_cached_pod_metadata(self, namespace_name, pod_name):
        self.logger.debug("Trying to find cached metadata for pod {}".format(pod_name))
        return self.__metadata[namespace_name][pod_name]

    def __query_pod_metadata_from_kubernetes_api_server(self, namespace_name, pod_name):
        self.logger.debug("Trying to query metadata for pod {}/{} from the Kubernetes API".format(namespace_name, pod_name))

        cached_namespace = self.__metadata.setdefault(namespace_name, {})
        cached_pod_metadata = cached_namespace.setdefault(pod_name, {})

        try:
            pod = self.__client_api.list_namespaced_pod(namespace_name, field_selector="metadata.name=={}".format(pod_name)).items[0]
        except (ApiException, IndexError) as e:
            self.logger.warning("Error querying pod status for pod {}/{}: {}".format(namespace_name, pod_name, e))
            return cached_pod_metadata

        try:
            cached_pod_metadata[self.add_prefix("pod_uuid")] = pod.metadata.uid
        except (AttributeError, IndexError, TypeError):
            self.logger.warning("Error querying pod.metadata.uid for pod {}/{} from the Kubernetes API".format(namespace_name, pod_name))

        for name, value in (pod.metadata.labels or {}).items():
            cached_pod_metadata[self.add_prefix("labels{}{}".format(self.key_delimiter, name))] = value

        for name, value in (pod.metadata.annotations or {}).items():
            cached_pod_metadata[self.add_prefix("annotations{}{}".format(self.key_delimiter, name))] = value

        try:
            container_status = pod.status.container_statuses[0]
            cached_pod_metadata[self.add_prefix("container_name")] = container_status.name
            cached_pod_metadata[self.add_prefix("container_image")] = container_status.image
            cached_pod_metadata[self.add_prefix("container_hash")] = container_status.image_id.replace("docker-pullable://", "", 1)
            cached_pod_metadata[self.add_prefix("docker_id")] = container_status.container_id.replace("docker://", "", 1)
        except (AttributeError, IndexError, TypeError):
            self.logger.warning("Error querying container_status for pod {}/{} from the Kubernetes API".format(namespace_name, pod_name))
        return cached_pod_metadata

    def get_pod_metadata(self, namespace_name, pod_name):
        try:
            return self.__get_cached_pod_metadata(namespace_name, pod_name)
        except KeyError:
            return self.__query_pod_metadata_from_kubernetes_api_server(namespace_name, pod_name)

    def parse(self, msg):
        namespace_name = msg[self.add_prefix("namespace_name")].decode()
        pod_name = msg[self.add_prefix("pod_name")].decode()

        pod_metadata = self.get_pod_metadata(namespace_name, pod_name)
        for name, value in pod_metadata.items():
            try:
                if not msg[name]:
                    msg[name] = value
            except KeyError:
                msg[name] = value

        return True



#!/usr/bin/env python3
#############################################################################
# Copyright (c) 2010-2017 Balabit
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################
#
# TODOs:
#  * be more clever about which rules to include as bison generates a lot of
#    warnings about unused rules
#

from __future__ import print_function
import fileinput
import os
import sys
import codecs

grammar_file = os.path.join(os.environ.get('top_srcdir', ''), 'lib/cfg-grammar.y')
if not os.path.isfile(grammar_file):
    grammar_file = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'cfg-grammar.y')
if not os.path.isfile(grammar_file):
    sys.exit('Error opening cfg-grammar.y')

def print_to_stdout(line):
    if sys.hexversion >= 0x3000000:
        sys.stdout.buffer.write(line.encode("utf-8"))
    else:
        print(line.encode("utf-8"), end='')

def include_block(block_type):
    start_marker = 'START_' + block_type
    end_marker = 'END_' + block_type

    with codecs.open(grammar_file, encoding="utf-8") as f:
        in_block = False
        for line in f:
            if start_marker in line:
                in_block = True
            elif end_marker in line:
                in_block = False
            elif in_block:
                print_to_stdout(line)

for line in fileinput.input(openhook=fileinput.hook_encoded("utf-8")):
    if 'INCLUDE_DECLS' in line:
        include_block('DECLS')
    elif 'INCLUDE_RULES' in line:
        include_block('RULES')
    else:
        print_to_stdout(line)



#############################################################################
# Copyright (c) 2015-2016 Balabit
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from __future__ import print_function, absolute_import
from .completerlang import CompleterLang
from .commandlinelexer import CommandLineLexer
from .getoptlexer import GetoptLexer


class TemplateFunctionLang(CompleterLang):
    tokens = []
    tokens_base = [
        "COMMAND", "ARG", "OPT"
    ]
    known_commands = ("format-json", "echo")
    known_options = (
        # value-pairs
        "--scope", "--exclude", "--key", "--rekey", "--pair",
        "--shift", "--add-prefix", "--replace-prefix", "--replace")

    def p_template_func(self, p):
        '''template_func : tf_format_json
                         | tf_echo
                         | tf_generic'''

    def p_tf_format_json(self, p):
        '''tf_format_json : COMMAND_FORMAT_JSON tf_format_json_args'''

    def p_tf_format_json_args(self, p):
        '''tf_format_json_args : tf_format_json_arg tf_format_json_args
                               | '''

    def p_tf_format_json_arg(self, p):
        '''tf_format_json_arg  : OPT__SCOPE value_pairs_scope
                               | OPT__EXCLUDE ARG
                               | OPT__KEY name_value_name
                               | OPT__REKEY ARG
                               | OPT__PAIR name_value_pair
                               | OPT__SHIFT ARG
                               | OPT__ADD_PREFIX ARG
                               | OPT__REPLACE_PREFIX ARG
                               | OPT__REPLACE ARG
                               | OPT
                               | ARG'''

    def p_name_value_name(self, p):
        '''name_value_name : ARG'''

    def p_name_value_pair(self, p):
        '''name_value_pair : ARG'''

    def p_value_pairs_scope(self, p):
        '''value_pairs_scope : ARG'''

    def p_tf_echo(self, p):
        '''tf_echo : COMMAND_ECHO templates'''

    def p_tf_generic(self, p):
        '''tf_generic : COMMAND args'''

    def p_args(self, p):
        '''args : ARG args
                |
        '''

    def p_templates(self, p):
        '''templates : template templates
                     | '''

    def p_template(self, p):
        '''template : ARG'''

    def p_ipaddress(self, p):
        '''ipaddress : ARG'''

    def _initialize_rules(self):
        tokens = list(self.tokens_base)
        self._convert_known_commands_to_tokens(tokens)
        self._convert_known_options_to_tokens(tokens)
        self.tokens = tokens

    def _construct_lexer(self):
        return GetoptLexer(CommandLineLexer(),
                           known_commands=self.known_commands,
                           known_options=self.known_options)

    @staticmethod
    def _tokenize_argument(arg):
        return arg.upper().replace('-', '_')

    def _convert_known_commands_to_tokens(self, tokens):
        for command in self.known_commands:
            tokens.append("COMMAND_{}".format(self._tokenize_argument(command)))

    def _convert_known_options_to_tokens(self, tokens):
        for option in self.known_options:
            tokens.append("OPT{}".format(self._tokenize_argument(option)))



#############################################################################
# Copyright (c) 2023 Attila Szakacs
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as published
# by the Free Software Foundation, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# As an additional exemption you are allowed to compile & link against the
# OpenSSL libraries as published by the OpenSSL project. See the file
# COPYING for details.
#
#############################################################################

from .s3_object import S3Object, S3ObjectPersist, SharedBool, ConstructorError, PersistLoadError, AlreadyFinishedError

try:
    from boto3 import client
    from botocore.exceptions import ClientError, EndpointConnectionError

    deps_installed = True
except ImportError:
    deps_installed = False

from concurrent.futures import ThreadPoolExecutor
from glob import glob
from logging import getLogger
from pathlib import Path
from signal import signal, SIGINT, SIG_IGN
from sys import exc_info
from syslogng import LogDestination, LogMessage, LogTemplate, get_installation_path_for
from threading import Event, Timer
from time import monotonic, sleep
from typing import Any, Dict, Optional

signal(SIGINT, SIG_IGN)


class S3Destination(LogDestination):
    S3_OBJECT_TIMEOUT_INTERVAL_SECONDS = 60

    def __init_options(self, options: Dict[str, Any]) -> None:
        try:
            self.url = str(options["url"])
            self.bucket = str(options["bucket"])
            self.access_key = str(options["access_key"])
            self.secret_key = str(options["secret_key"])
            self.object_key: LogTemplate = options["object_key"]
            self.object_key_timestamp: Optional[LogTemplate] = options["object_key_timestamp"]
            self.message_template: LogTemplate = options["template"]
            self.compression = bool(options["compression"])
            self.compresslevel = int(options["compresslevel"])
            self.chunk_size = int(options["chunk_size"])
            self.max_object_size = int(options["max_object_size"])
            self.upload_threads = int(options["upload_threads"])
            self.max_pending_uploads = int(options["max_pending_uploads"])
            self.flush_grace_period = int(options["flush_grace_period"])
            self.region: Optional[str] = str(options["region"])
            self.storage_class = str(options["storage_class"]).upper().replace("-", "_")
            self.canned_acl = str(options["canned_acl"]).lower().replace("_", "-")
        except KeyError:
            assert False, (
                f"S3: {str(exc_info()[1])[1:-1]}() option is missing. "
                "If you are using this driver via the python() destination, please use the s3() driver directly. "
                "The option validation and propagation should be done by the s3.conf SCL."
            )

        if str(self.object_key_timestamp) == "":
            self.object_key_timestamp = None

        if self.compresslevel < 0 or self.compresslevel > 9:
            self.logger.warn("compresslevel() must be an integer between 0 and 9. Using 9")
            self.compresslevel = 9

        if self.chunk_size < S3Object.MIN_CHUNK_SIZE_BYTES:
            self.logger.warn(f"chunk-size() must be at least {S3Object.MIN_CHUNK_SIZE_BYTES}. Using minimal value")
            self.chunk_size = S3Object.MIN_CHUNK_SIZE_BYTES

        if self.upload_threads < 1:
            self.logger.warn("upload-threads() must be a positive integer. Using 1")
            self.upload_threads = 1

        if self.max_pending_uploads < 1:
            self.logger.warn("max-pending-uploads() must be a positive integer. Using 1")
            self.max_pending_uploads = 1

        if self.flush_grace_period < 1:
            self.logger.warn("flush-grace-period() must be a positive integer. Using 1")
            self.flush_grace_period = 1

        if self.region == "":
            self.region = None

        VALID_STORAGE_CLASSES = {
            "STANDARD",
            "REDUCED_REDUNDANCY",
            "STANDARD_IA",
            "ONEZONE_IA",
            "INTELLIGENT_TIERING",
            "GLACIER",
            "DEEP_ARCHIVE",
            "OUTPOSTS",
            "GLACIER_IR",
            "SNOW",
        }
        if self.storage_class not in VALID_STORAGE_CLASSES:
            self.logger.warn(
                f"Invalid storage-class(). Valid values are: f{', '.join(sorted(VALID_STORAGE_CLASSES))}. Using STANDARD"
            )
            self.storage_class = "STANDARD"

        VALID_CANNED_ACLS = {
            "private",
            "public-read",
            "public-read-write",
            "aws-exec-read",
            "authenticated-read",
            "bucket-owner-read",
            "bucket-owner-full-control",
            "log-delivery-write",
        }
        if self.canned_acl != "" and self.canned_acl not in VALID_CANNED_ACLS:
            self.logger.warn(
                f"Invalid canned-acl(). Valid values are: f{', '.join(sorted(VALID_CANNED_ACLS))} or empty. Using empty"
            )
            self.canned_acl = ""

    def init(self, options: Dict[str, Any]) -> bool:
        self.logger = getLogger("S3")

        if not deps_installed:
            self.logger.error(
                "Unable to start the Python based S3 destination, the required Python dependencies (`boto3` and/or `botocore`) are missing"
            )
            return False

        self.client: Optional[Any] = None

        self.s3_objects: Dict[str, S3Object] = dict()
        self.finished_s3_objects: Dict[str, Dict[str, Dict[int, S3Object]]] = dict()
        self.backfill_s3_objects: Dict[str, Dict[str, S3Object]] = dict()

        self.__init_options(options)

        self.exit_requested = Event()
        self.executor = ThreadPoolExecutor(max_workers=self.upload_threads)
        self.flush_poll_event: Optional[Timer] = None

        self.working_dir = Path(get_installation_path_for(r"${localstatedir}"), "s3")
        self.persist_loaded = False

        return True

    def deinit(self) -> None:
        self.executor.shutdown()

    @staticmethod
    def generate_persist_name(options: Dict[str, Any]) -> str:
        return f"s3({','.join([options['url'], options['bucket'], str(options['object_key'])])})"

    def __load_persist(self) -> None:
        for path_str in glob(pathname="*.json", root_dir=self.working_dir):
            path = Path(self.working_dir, path_str)
            try:
                persist = S3ObjectPersist.load(path=path)
            except PersistLoadError:
                self.logger.error(f"Cannot load persist file: {str(path)}")
                continue

            if persist.persist_name != self.persist_name:
                continue

            try:
                s3_object = S3Object.load_finished(
                    persist=persist,
                    executor=self.executor,
                    client=self.client,
                    logger=self.logger,
                    exit_requested=self.exit_requested,
                )
            except ConstructorError as e:
                self.logger.error(str(e))
                continue

            assert s3_object.finished

            target_key = s3_object.target_key
            self.finished_s3_objects.setdefault(target_key, dict()).setdefault(s3_object.timestamp, dict())[
                s3_object.index
            ] = s3_object

            # Completely new target
            if target_key not in self.s3_objects:
                self.s3_objects[target_key] = s3_object
                continue

            # Existing target
            existing_s3_object = self.s3_objects[target_key]

            # Same target and timestamp as stored
            if existing_s3_object.timestamp == s3_object.timestamp:
                if existing_s3_object.index < s3_object.index:
                    self.s3_objects[target_key] = s3_object
                continue

            # Target for a fresher timestamp
            if existing_s3_object.timestamp < s3_object.timestamp:
                self.s3_objects[target_key] = s3_object
                continue

            # Target for an older timestamp
            continue

        self.persist_loaded = True

    def __create_bucket(self) -> bool:
        assert self.client
        try:
            self.client.create_bucket(Bucket=self.bucket)
        except (ClientError, EndpointConnectionError) as e:
            self.logger.error(f"Failed to create bucket ({self.bucket}): {e}")
            return False

        self.logger.info(f"Bucket ({self.bucket}) successfully created")
        return True

    def open(self) -> bool:
        if self.is_opened():
            return True

        self.client = client(
            service_name="s3",
            endpoint_url=self.url,
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name=self.region,
        )

        is_opened = False
        try:
            self.client.head_bucket(Bucket=self.bucket)
            is_opened = True
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                self.logger.info(f"Bucket ({self.bucket}) does not exist, trying to create it")
                is_opened = self.__create_bucket()
        except EndpointConnectionError:
            pass

        if not is_opened:
            self.client = None
            return False

        if not self.persist_loaded:
            self.__load_persist()

        self.exit_requested.clear()

        if self.flush_poll_event is None:
            self.__start_flush_poll_timer()

        return True

    def is_opened(self) -> bool:
        return self.client is not None

    def close(self) -> None:
        if not self.is_opened():
            return

        assert self.client is not None

        self.exit_requested.set()

        if self.flush_poll_event is not None:
            self.flush_poll_event.cancel()
            self.flush_poll_event = None

        for s3_object in self.s3_objects.values():
            if not s3_object.finished:
                self.__finish_s3_object(s3_object)

        for backfill_timestamps in self.backfill_s3_objects.values():
            for s3_object in backfill_timestamps.values():
                if not s3_object.finished:
                    self.__finish_s3_object(s3_object)

        for finished_timestamps in self.finished_s3_objects.values():
            for indexes in finished_timestamps.values():
                for s3_object in indexes.values():
                    s3_object.wait_for_upload_to_complete()

        self.finished_s3_objects.clear()

        self.client.close()
        self.client = None

    def __start_flush_poll_timer(self) -> None:
        self.flush_poll_event = Timer(
            interval=self.S3_OBJECT_TIMEOUT_INTERVAL_SECONDS, function=self.__flush_timed_out_s3_objects
        )
        self.flush_poll_event.start()

    def __flush_timed_out_s3_objects(self) -> None:
        def should_flush_s3_object(s3_object: S3Object, now: float) -> bool:
            return not s3_object.finished and s3_object.modified_at + self.flush_grace_period * 60 <= now

        now = monotonic()

        for s3_object in self.s3_objects.values():
            if should_flush_s3_object(s3_object, now):
                self.__finish_s3_object(s3_object)

        for timestamps in self.backfill_s3_objects.values():
            for s3_object in timestamps.values():
                if should_flush_s3_object(s3_object, now):
                    self.__finish_s3_object(s3_object)

        self.__start_flush_poll_timer()

    def __format_target_key(self, msg: LogMessage) -> str:
        return self.object_key.format(msg, self.template_options)

    def __format_timestamp(self, msg: LogMessage) -> str:
        if self.object_key_timestamp is not None:
            return self.object_key_timestamp.format(msg, self.template_options)
        return ""

    def __create_initial_s3_object(self, target_key: str, timestamp: str) -> S3Object:
        return S3Object.create_initial(
            working_dir=self.working_dir,
            bucket=self.bucket,
            target_key=target_key,
            timestamp=timestamp,
            compress=self.compression,
            storage_class=self.storage_class,
            persist_name=self.persist_name,
            executor=self.executor,
            compresslevel=self.compresslevel,
            chunk_size=self.chunk_size,
            canned_acl=self.canned_acl,
            client=self.client,
            logger=self.logger,
            exit_requested=self.exit_requested,
        )

    def __finish_s3_object(self, s3_object: S3Object) -> None:
        s3_object.finish()
        d = self.finished_s3_objects.setdefault(s3_object.target_key, dict()).setdefault(s3_object.timestamp, dict())
        d[s3_object.index] = s3_object

    def __get_backfill_s3_object(self, target_key: str, timestamp: str) -> S3Object:
        backfill_target_s3_objects = self.backfill_s3_objects.setdefault(target_key, dict())

        # Completely new backfill
        if timestamp not in backfill_target_s3_objects:
            try:
                # Backfill for a previously finished target
                indexes = list(self.finished_s3_objects[target_key][timestamp].keys())
                s3_object = self.finished_s3_objects[target_key][timestamp][max(indexes)].create_next()
            except KeyError:
                # Backfill for an unprocessed target
                s3_object = self.__create_initial_s3_object(target_key, timestamp)
            backfill_target_s3_objects[timestamp] = s3_object
            return s3_object

        # Existing backfill
        s3_object = backfill_target_s3_objects[timestamp]

        if not s3_object.finished:
            return s3_object

        s3_object = backfill_target_s3_objects[timestamp] = s3_object.create_next()
        return s3_object

    def __get_s3_object(self, msg: LogMessage) -> S3Object:
        target_key = self.__format_target_key(msg)
        timestamp = self.__format_timestamp(msg)

        # Completely new target
        if target_key not in self.s3_objects:
            s3_object = self.s3_objects[target_key] = self.__create_initial_s3_object(target_key, timestamp)
            return s3_object

        # Existing target
        s3_object = self.s3_objects[target_key]

        # Same target and timestamp as stored
        if s3_object.timestamp == timestamp:
            if not s3_object.finished:
                return s3_object

            s3_object = self.s3_objects[target_key] = s3_object.create_next()
            return s3_object

        # Target for a fresher timestamp
        if s3_object.timestamp < timestamp:
            self.__finish_s3_object(s3_object)
            s3_object = self.s3_objects[target_key] = self.__create_initial_s3_object(target_key, timestamp)
            return s3_object

        # Target for an older timestamp
        return self.__get_backfill_s3_object(target_key, timestamp)

    def __ensure_free_space_in_executor_queue(self) -> bool:
        """Waits for 1 minute."""
        retries = 0

        while self.executor._work_queue.qsize() >= self.max_pending_uploads:
            if retries >= 600:
                self.logger.error("Upload queue is still full after 1 minute, reconnecting")
                return False

            if retries == 0:
                self.logger.info("Upload queue is full, waiting for available space")

            sleep(0.1)
            retries += 1

        return True

    def send(self, msg: LogMessage) -> int:
        if not self.__ensure_free_space_in_executor_queue():
            return self.NOT_CONNECTED

        try:
            s3_object = self.__get_s3_object(msg)
        except ConstructorError as e:
            self.logger.error(str(e))
            return self.NOT_CONNECTED

        data = self.message_template.format(msg, self.template_options).encode("utf-8")

        try:
            s3_object.write(data)
        except OSError as e:
            self.logger.error(f"Failed to write data: {e}")
            return self.ERROR
        except AlreadyFinishedError as e:
            self.logger.error(f"Failed to write data: {e}")
            self.__finish_s3_object(s3_object)
            return self.ERROR

        if s3_object.finished:
            # The S3 object finished itself after a successful write.
            self.__finish_s3_object(s3_object)

        if s3_object.size >= self.max_object_size:
            self.__finish_s3_object(s3_object)

        return self.SUCCESS
